import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
part 'extension.dart';

enum DEVICE {
  MOBILE_PORTRAIT,
  TABLET_PORTRAIT,
  WEB, // web is equal to mobile landscape and tablet landscape
}

// ignore: avoid_classes_with_only_static_members
/// Class use for beautiful responsive design both on tablets and mobiles
abstract class ResponsiveSizeUtils {
  static late double _designedDeviceWidth;
  static late double _designedDeviceHeight;

  static late double _screenWidth;
  static late double _screenHeight;
  static late Orientation _orientation;

  static void _calculate({
    required double screenWidth,
    required double screenHeight,
    required Orientation orientation,
  }) {
    _screenWidth = screenWidth;
    _screenHeight = screenHeight;
    _orientation = orientation;
  }

  static void _initDesignedSize({
    required double designedDeviceWidth,
    required double designedDeviceHeight,
  }) {
    _designedDeviceWidth = designedDeviceWidth;
    _designedDeviceHeight = designedDeviceHeight;
  }

  /// return a reasonable width `value`.
  ///
  /// if `maxWidth` is null then the `value` will be at most `designedWidth`
  /// logical pixel otherwise, the `value` will be at most `maxWidth` logical pixels.
  ///
  /// `value` can be any double between zero and maximum width.
  ///
  static double width(
    double designedWidth, {
    double? maxWidth,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _width = _screenWidth * designedWidth / _designedDeviceWidth;
      if (maxWidth == null) {
        if (_width <= designedWidth) {
          return _width;
        } else {
          return designedWidth;
        }
      } else {
        return _comparator(
          value: designedWidth,
          maxValue: maxWidth,
        );
      }
    } else {
      final _width = _screenWidth * designedWidth / _designedDeviceHeight;
      if (maxWidth == null) {
        if (_width <= designedWidth) {
          return _width;
        } else {
          return designedWidth;
        }
      } else {
        return _comparator(
          value: designedWidth,
          maxValue: maxWidth,
        );
      }
    }
  }

  /// return a reasonable height `value`.
  ///
  /// if `maxHeight` is null then the `value` will be at most `designedHeight`
  /// logical pixel otherwise, the `value` will be at most `maxHeight` logical pixels.
  ///
  /// `value` can be any double between zero and maximum height.
  ///
  static double height(
    double designedHeight, {
    double? maxHeight,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _height = _screenHeight * designedHeight / _designedDeviceHeight;
      if (maxHeight == null) {
        if (_height <= designedHeight) {
          return _height;
        } else {
          return designedHeight;
        }
      } else {
        return _comparator(
          value: _height,
          maxValue: maxHeight,
        );
      }
    } else {
      final _height = _screenHeight * designedHeight / _designedDeviceWidth;
      if (maxHeight == null) {
        if (_height <= designedHeight) {
          return _height;
        } else {
          return designedHeight;
        }
      } else {
        return _comparator(
          value: _height,
          maxValue: maxHeight,
        );
      }
    }
  }

  /// return `width` of screen
  static double fullWidth() {
    return _screenWidth;
  }

  /// return `height` of screen
  static double fullHeight() {
    return _screenHeight;
  }

  static double? widthSpacing({
    required double spaceSizeBeforeBreakPoint,
    required double spaceSizeAfterBreakPoint,
  }) {
    final _screenWidth = ResponsiveSizeUtils._screenWidth;
    return _screenWidth <= _designedDeviceWidth
        ? width(
            spaceSizeBeforeBreakPoint,
          )
        : width(
            spaceSizeAfterBreakPoint,
          );
  }

  static double heightSpacing({
    required double spaceSizeBeforeBreakPoint,
    required double spaceSizeAfterBreakPoint,
  }) {
    return _screenHeight <= _designedDeviceHeight
        ? height(
            spaceSizeBeforeBreakPoint,
          )
        : height(
            spaceSizeAfterBreakPoint,
          );
  }

  ///
  static double widthFraction({
    required double fraction,
  }) {
    assert(
      fraction < 1,
      'fraction must be lower than 1',
    );
    return _screenWidth * fraction;
  }

  static double heightFraction({
    required double fraction,
  }) {
    assert(
      fraction < 1,
      'fraction must be lower than 1',
    );
    return _screenHeight * fraction;
  }

  /// Show if device is in portrait mode
  static bool isPortrait() {
    return _orientation == Orientation.portrait;
  }

  /// Return smallest width of device
  static double smallestWidth() {
    final _width = _screenWidth;
    final _height = _screenHeight;
    return _width <= _height ? _width : _height;
  }

  /// Return [DEVICE] type
  /// 0 <= [Mobile portrait] <= 480 <= [Tablet portrait] <= 720 <= [Web]
  /// [Web] is also equal to [mobile landscape] and [tablet landscape] and also [desktop]
  static DEVICE device() {
    final _smallestWidth = smallestWidth();
    if (_smallestWidth <= 480.0) {
      return DEVICE.MOBILE_PORTRAIT;
    } else if (_smallestWidth <= 720.0) {
      return DEVICE.TABLET_PORTRAIT;
    } else {
      return DEVICE.WEB;
    }
  }

  static double _comparator({
    required double value,
    required double maxValue,
  }) {
    return value <= maxValue ? value : maxValue;
  }

  /// this widget gives you three layout so that you can have different widget
  /// for each device
  static Widget responsiveWidgetAccordingToDevice({
    required Widget mobilePortrait,
    required Widget tabletPortrait,
    required Widget web,
  }) {
    switch (device()) {
      case DEVICE.WEB:
        return web;
      case DEVICE.TABLET_PORTRAIT:
        return tabletPortrait;
      default:
        return mobilePortrait;
    }
  }

  static double responsiveFont({
    required double fontSize,
    double? maxFontSize,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _calculatedFontSize =
          _screenHeight * fontSize / _designedDeviceHeight;
      if (maxFontSize == null) {
        if (_calculatedFontSize <= fontSize) {
          return _calculatedFontSize;
        } else {
          return fontSize;
        }
      } else {
        return _comparator(
          value: _calculatedFontSize,
          maxValue: maxFontSize,
        );
      }
    } else {
      final _calculatedFontSize =
          _screenHeight * fontSize / _designedDeviceWidth;
      if (maxFontSize == null) {
        if (_calculatedFontSize <= fontSize) {
          return _calculatedFontSize;
        } else {
          return fontSize;
        }
      } else {
        return _comparator(
          value: _calculatedFontSize,
          maxValue: maxFontSize,
        );
      }
    }
  }
}

class ResponsiveWidget extends StatelessWidget {
  /// A helper class that init the ResponsiveSizeUtils class
  /// wrap build method of first widget
  ResponsiveWidget({
    Key? key,
    required this.child,
    required this.designedSize,
    this.onInit,
  }) : super(key: key) {
    ResponsiveSizeUtils._initDesignedSize(
      designedDeviceWidth: designedSize.width,
      designedDeviceHeight: designedSize.height,
    );
    if (onInit != null) {
      onInit!.call();
    }
  }

  final Widget child;
  final Size designedSize;
  final void Function()? onInit;

  @override
  Widget build(
    BuildContext context,
  ) {
    return MaterialApp(
      home: OrientationBuilder(
        builder: (
          context,
          orientation,
        ) {
          return LayoutBuilder(
            builder: (
              context,
              constraints,
            ) {
              if (orientation == Orientation.portrait) {
                ResponsiveSizeUtils._calculate(
                  screenWidth: constraints.maxWidth,
                  screenHeight: constraints.maxHeight,
                  orientation: orientation,
                );
              } else {
                ResponsiveSizeUtils._calculate(
                  screenWidth: constraints.maxHeight,
                  screenHeight: constraints.maxWidth,
                  orientation: orientation,
                );
              }
              return child;
            },
          );
        },
      ),
    );
  }
}
